/* cursor movement and output buffering functions */

#include <stdio.h>
#include <sys/ioctl.h>

#ifndef BSD
#include <termio.h>
#else
#include <sgtty.h>
#endif

#define STDOUT  1       /* standard output file descriptor */

extern  char    *argv0; /* command name (must be set in main function) */

/* termcap library global data needed by tgoto() */
#if defined(SV_R1) | defined(BSD)
extern  char    *BC;    /* backspace character sequence if not control-h */
extern  char    PC;     /* pad character (other than null) */
extern  char    *UP;    /* cursor up character sequence */
#endif

int     CO;             /* number of columns in a line */
int     LI;             /* number of lines on screen */
char    *term;          /* terminal type */

static  char    *CE;            /* clear to end of line character sequence */
static  char    *CL;            /* clear screen character sequence */
static  char    *CM;            /* cursor motion character sequence */
static  char    *TE;            /* cursor motion ending character sequence */
static  char    *TI;            /* cursor motion initialization sequence */
static  char    chars[128];     /* termcap character sequence storage */
static  char    *charsp;        /* termcap character sequence free space pointer */
static  char    outbuf[BUFSIZ]; /* terminal output buffer */
static  int     buffer = 0;     /* buffer semaphore */

char    *getenv(), *tgetstr(), *tgoto();
int     fputchar();             /* needed by fputs() */
void    at(), exit();

/* cursor motion initialization */

void
curs_init() 
{
        char    bp[1024];

        /* get the terminal type  */
        if ((term = getenv("TERM")) == NULL) {
                fprintf(stderr, "%s: TERM variable is not set or is not exported in your .profile\n", argv0);
                exit(1);
        }
        /* get the terminal data base entry */
        if (tgetent(bp, term) < 0) {
                fprintf(stderr, "%s: \"%s\" is not in the terminal data base\n",
                        argv0, term);
                exit(1);
        }
        /* get the terminal attributes */
        charsp = chars;
#if defined(SV_R1) | defined(BSD)
        BC = tgetstr("bc", &charsp);    /* backspace character if not ^H */
        PC = tgetflag("pc");            /* pad character if not '\0' */
        UP = tgetstr("up", &charsp);    /* cursor up */
#endif
        CE = tgetstr("ce", &charsp);    /* clear to end of line */
        CL = tgetstr("cl", &charsp);    /* clear screen */
        CM = tgetstr("cm", &charsp);    /* cursor motion */
        CO = tgetnum("co");             /* number of columns in a line */
        LI = tgetnum("li");             /* number of lines on screen */
        TE = tgetstr("te", &charsp);    /* cursor cleanup */
        TI = tgetstr("ti", &charsp);    /* cursor initialization */
        
        /* verify that this terminal can be used */
        if (LI < 0) {
                fprintf(stderr, "%s: \"%s\" is not a screen terminal\n", 
                        argv0, term);
                exit(1);
        }
        else if (CM == (char *) NULL) {
                fprintf(stderr, "%s: \"%s\" is not a cursor addressable terminal\n",
                        argv0, term);
                exit(1);
        }
}

/* begin cursor motion */

void
curs_begin()
{
        tputs(TI, 0, fputchar);
        fflush(stdout);         /* System V line buffers to terminal */
}

/* end cursor motion */

void
curs_end()
{
        tputs(TE, 0, fputchar);
        fflush(stdout);         /* System V line buffers to terminal */
}

/* erase current line */

void
clearline() 
{
        /* see if this terminal can clear to the end of the line */
        if (CE != NULL) {
                tputs(CE, 0, fputchar);
        }
        else {
                printf("\r%*s\r", CO - 1, "");
        }
}

/* clear the screen */

void
clearscreen() 
{
        /* abort output so the user does not have to wait */
        fflush(stdout);
#ifdef BSD
        ioctl(STDOUT, TIOCFLUSH, 0);
#else
        ioctl(STDOUT, TCFLSH, 1);
#endif
        
        /* resync the terminal */
        putchar('\r');
        at(1, 1);

        /* clear the screen */
        tputs(CL, 0, fputchar);
}

/* position cursor at requested row and column */

void
at(row, column)  
register row, column;
{
        tputs(tgoto(CM, column - 1, row - 1), 0, fputchar);
}

/* tputs() requires a function to put a character, and putchar() is a macro */

int
fputchar(c)
{
        putchar(c);
}

/* buffer output for speed */

void
startbuffering() 
{
        if (buffer++ == 0) {
                fflush(stdout); /* SV_R2 buffers to terminal */
                setbuf(stdout, outbuf);
        }
}

/* stop buffering output */

void
stopbuffering() 
{
        if (--buffer == 0) {
                fflush(stdout);
                setbuf(stdout, (char *) NULL);
        }
}
