/*      cscope - interactive C symbol or text cross-reference
 *
 *      searching functions
 */

/* most of these functions have been optimized so their innermost loops have
 * only one test for the desired character by putting the char and 
 * an end-of-block marker (\0) at the end of the disk block buffer.
 * When the inner loop exits on the char, an outer loop will see if
 * the char is followed by a \0.  If so, it will read the next block
 * and restart the inner loop.
 */

#include "strings.h"                    /* CNULL */
#include "global.h"

char    *blockp;                        /* pointer to current char in block */

static  char    block[BUFSIZ + 2];      /* leave room for end-of-block mark */
static  int     blocklen;               /* length of disk block read */
static  char    mark;                   /* mark character to be searched for */

void    putline();
BOOL    findfunction(), match();
/* find the symbol in the cross-reference */

void
findsymbol()
{
        char    file[PATHLEN + 1];      /* source file name */
        char    function[PATLEN + 1];   /* function name */
        char    symbol[PATLEN + 1];     /* symbol name */
        char    *s;
        register char   *cp;

        /* skip the header to avoid matching a file name */
        cp = readblock();       /* read the first cross-reference block */
        scanpast('\t');         /* find the end of the header */
        nextchar();             /* skip the file marker */
        putstring(file);        /* save the file name */
        cp = blockp + 1;        /* skip the \n */
        setmark('\n');          /* set the mark (search) character */
        
        /* while not end-of-block or not end-of-file */
        while (*cp != '\n' || (cp = readblock()) != NULL) {

                /* look for a source file or function name */
                if (*cp == '\t') {
                        blockp = cp;
                        switch (*(nextchar())) {
                                
                        case NEWFILE:           /* file name */
                                s = file;
                                progress();
                                break;
                        case FCNDEF:            /* function name */
                                s = function;
                                break;
                        default:                /* other symbol */
                                s = symbol;
                        }
                        /* save the name and compare it to the pattern */
                        nextchar();
                        putstring(s);
                        if (strcmp(s, pattern) == 0 && s != file) {
                                goto matched;
                        }
                        setmark('\n');          /* restore the \n mark */
                        cp = blockp + 1;        /* skip the \n */
                        continue;
                }
                /* match the symbol to the pattern */
                if (*cp++ != pattern[0]) { /* quick check for efficiency */
                        goto nomatch;
                }
                /* this is executed on about 1 out of 20 symbols */
                blockp = cp;
                if (match(1)) {
        matched:                
                        /* output the file and function names */
                        fprintf(refsfound, "%s %s ", file, function);
                        
                        /* output the source line */
                        scanpast(' ');
                        putline('\n');
                        setmark('\n');          /* restore the \n mark */
                        cp = blockp + 1;        /* skip the \n */
                }
                else {
                        cp = blockp;            /* next char could be \n */
        nomatch:
                        /* skip to the \n */
                        do {    /* innermost loop optimized to only one test */
                                while (*cp++ != '\n') {
                                        ;
                                }
                        } while (*cp == '\0' && (cp = readblock()) != NULL);
                }
        }
}
/* find the function definition or #define */

void
finddef()
{
        char    file[PATHLEN + 1];      /* source file name */

        /* find the function definition */
        if (findfunction(file) == YES) {

                /* output the file and function name */
                fprintf(refsfound, "%s %s ", file, pattern);

                /* output the source line */
                scanpast(' ');
                putline('\n');
        }
}
/* find the functions calling this function */

void
calling()
{
        char    file[PATHLEN + 1];      /* source file name */
        char    function[PATLEN + 1];   /* function name */

        readblock();    /* read the first cross-reference block */
        
        /* find the next file name or function definition */
        while (scanpast('\t') != NULL) {
                switch (*blockp) {
                        
                case NEWFILE:           /* save file name */
                        nextchar();
                        putstring(file);
                        progress();
                        break;
                        
                case FCNDEF:            /* save calling function name */
                        nextchar();
                        putstring(function);
                        break;
                        
                case FCNCALL:           /* match function called to pattern */
                        nextchar();
                        if (match(0)) {
                                
                                /* output the file and calling function name */
                                fprintf(refsfound, "%s %s ", file, function);

                                /* output the source line */
                                scanpast(' ');
                                putline('\n');
                        }
                }
        }
}
/* find the functions called by this function */

void
calledby()
{
        char    file[PATHLEN + 1];      /* source file name */

        /* find the function definition */
        if (findfunction(file) == YES) {

                /* find the next function call or the end of this function */
                while (scanpast('\t') != NULL) {
                        switch (*blockp) {
                        
                        case FCNCALL:           /* function call */

                                /* output the file and function name */
                                fprintf(refsfound, "%s ", file);
                                nextchar();
                                putline(' ');

                                /* output the source line */
                                scanpast(' ');
                                putline('\n');
                                break;
                        
                        case FCNDEF:            /* function end */
                                return;
                        }
                }
        }
}
/* find the function definition or #define in the cross-reference */

static  BOOL
findfunction(file)
char    *file;
{
        readblock();    /* read the first cross-reference block */
        
        /* find the next file name or definition */
        while (scanpast('\t') != NULL) {
                switch (*blockp) {
                        
                case NEWFILE:           /* save file name */
                        nextchar();
                        putstring(file);
                        progress();
                        break;

                case DEFINE:
                case FCNDEF:            /* match name to pattern */
                        nextchar();
                        if (match(0)) {
                                return(YES);
                        }
                        break;
                }
        }
        return(NO);
}
/* match the pattern to the name */

static  BOOL
match(i)
int     i;
{
        do {
                while (*blockp == pattern[i]) {
                        ++blockp;
                        ++i;
                }
        } while (*(blockp + 1) == '\0' && readblock() != NULL);
        
        if (*blockp == '\n' && pattern[i] == '\0') {
                return(YES);
        }
        return(NO);
}

/* put the rest of the cross-reference line into the references found file */

static  void
putline(separator)
char    separator;
{
        register char   *cp;
        
        setmark('\n');
        cp = blockp;
        do {
                while (*cp != '\n') {
                        putc(*cp++, refsfound);
                }
        } while (*(cp + 1) == '\0' && (cp = readblock()) != NULL);
        blockp = cp;
        putc(separator, refsfound);
}

/* put the rest of the cross-reference line into the string */

void
putstring(s)
register char   *s;
{
        register char   *cp;
        
        setmark('\n');
        cp = blockp;
        do {
                while (*cp != '\n') {
                        *s++ = *cp++;
                }
        } while (*(cp + 1) == '\0' && (cp = readblock()) != NULL);
        blockp = cp;
        *s = '\0';
}
/* scan past the next occurence of this character in the cross-reference */

char    *
scanpast(c)
register char   c;
{
        register char   *cp;
        
        setmark(c);
        cp = blockp;
        do {    /* innermost loop optimized to only one test */
                while (*cp != c) {
                        ++cp;
                }
        } while (*(cp + 1) == '\0' && (cp = readblock()) != NULL);
        blockp = cp;
        nextchar();     /* skip the found character */
        return(blockp);
}

/* move forward a character in the cross-reference */

char    *
nextchar()
{
        if (blockp != NULL) {   /* if not at end-of-file */
                if (*(++blockp + 1) == '\0') {  /* check for end-of-block */
                        blockp = readblock();
                }
        }
        return(blockp);
}

/* set the mark character for searching */

setmark(c)
char    c;
{
        mark = c;
        block[blocklen] = mark;
}

/* read a block of the cross-reference */

char    *
readblock()
{
        /* read the next block */
        blocklen = read(symrefs, block, BUFSIZ);
        blockp = block;
        
        /* add the search character and end-of-block mark */
        block[blocklen] = mark;
        block[blocklen + 1] = '\0';
        
        /* return NULL on end-of-file */
        if (blocklen == 0) {
                return(NULL);
        }
        return(block);
}
/* find the text in the source files */

void
findstring()
{
        char    greppat[2 * PATLEN];
        char    *cp, *pp;
        register int    i;

        /* translate special characters in the regular expression */
        cp = greppat;
        for (pp = pattern; *pp != '\0'; ++pp) {
                if (strchr(".*[\\^$", *pp) != NULL) {
                        *cp++ = '\\';
                }
                *cp++ = *pp;
        }
        *cp = '\0';
        
        /* search the source files */
        grepinit(greppat, refsfound, "%s <unknown> %ld ");
        for (i = 0; i < nsrcfiles; ++i) {
                progress();
                if (grep(srcfiles[i]) < 0) {
                        startbuffering();
                        at(2, 1);
                        clearline();
                        printf("Cannot open file %s", srcfiles[i]);
                        stopbuffering();
                }
        }
}

/* find file names matching the text */

void
findfile()
{
        register int    i;
        
        _str2pat(pattern);      /* calculate the pattern table once */
        for (i = 0; i < nsrcfiles; ++i) {
                if (strfind(srcfiles[i], NULL) != NULL) {
                        fprintf(refsfound, "%s <unknown> 1 <unknown>\n", 
                                srcfiles[i]);
                }
        }
}
